@@ -4,33 +4,41 @@ import builtins
 import os
 import re
 import shutil
-from textwrap import dedent, indent
+from dataclasses import dataclass, field
+from itertools import chain
 from pathlib import Path
+from textwrap import dedent, indent, wrap
 from typing import (
+    Any,
+    Dict,
     Generator,
     Iterable,
-    List,
-    Dict,
-    Set,
-    Union,
     Iterator,
-    Tuple,
+    List,
     Optional,
+    Set,
+    Tuple,
+    Union,
 )
-from dataclasses import dataclass
 
 import black
-import isort
+import isort.api
+from autoflake import fix_code
+from numpydoc.docscrape import NumpyDocString, Parameter
 from xmlschema import XMLSchema, qnames
 from xmlschema.validators import (
     XsdAnyAttribute,
     XsdAnyElement,
     XsdAttribute,
+    XsdComponent,
     XsdElement,
     XsdType,
-    XsdComponent,
 )
 
+# Track all camel-to-snake and pluralization results so we can include them in the model.
+camel_snake_registry: Dict[str, str] = {}
+plural_registry: Dict[Tuple[str, str], str] = {}
+
 
 # FIXME: Work out a better way to implement these override hacks.
 
@@ -49,6 +57,23 @@ class Override:
             self.body = indent(dedent(self.body), " " * 4)
 
 
+@dataclass
+class ClassOverride:
+    base_type: Optional[str] = None
+    imports: Optional[str] = None
+    fields: Optional[str] = None
+    fields_suppress: Set[str] = field(default_factory=set)
+    body: Optional[str] = None
+
+    def __post_init__(self) -> None:
+        if self.imports:
+            self.imports = dedent(self.imports)
+        if self.fields:
+            self.fields = indent(dedent(self.fields), " " * 4)
+        if self.body:
+            self.body = indent(dedent(self.body), " " * 4)
+
+
 # Maps XSD TypeName to Override configuration, used to control output for that type.
 OVERRIDES = {
     "MetadataOnly": Override(type_="bool", default="False"),
@@ -166,6 +191,7 @@ OVERRIDES = {
             from .file_annotation import FileAnnotation
             from .list_annotation import ListAnnotation
             from .long_annotation import LongAnnotation
+            from .map_annotation import MapAnnotation
             from .tag_annotation import TagAnnotation
             from .term_annotation import TermAnnotation
             from .timestamp_annotation import TimestampAnnotation
@@ -178,6 +204,7 @@ OVERRIDES = {
                 "file_annotation": FileAnnotation,
                 "list_annotation": ListAnnotation,
                 "long_annotation": LongAnnotation,
+                "map_annotation": MapAnnotation,
                 "tag_annotation": TagAnnotation,
                 "term_annotation": TermAnnotation,
                 "timestamp_annotation": TimestampAnnotation,
@@ -220,15 +247,67 @@ OVERRIDES = {
                 value: UniversallyUniqueIdentifier
         """,
     ),
+    "M/K": Override(type_="str", default=""),
 }
 
 
-def black_format(text: str, line_length: int = 79) -> str:
+# Maps XSD TypeName to ClassOverride configuration, used to control dataclass
+# generation.
+CLASS_OVERRIDES = {
+    "OME": ClassOverride(
+        imports="""
+            from typing import Any
+            import weakref
+            from ome_types import util
+        """,
+        body="""
+            def __post_init_post_parse__(self: Any, *args: Any) -> None:
+                ids = util.collect_ids(self)
+                for ref in util.collect_references(self):
+                    ref.ref_ = weakref.ref(ids[ref.id])
+        """,
+    ),
+    "Reference": ClassOverride(
+        imports="""
+            from dataclasses import field
+            from typing import Any, Optional
+            from .simple_types import LSID
+        """,
+        # FIXME Figure out typing for ref_ (weakref). Even with the "correct"
+        # typing, Pydantic has a problem.
+        fields="""
+            id: LSID
+            ref_: Any = field(default=None, init=False)
+        """,
+        # FIXME Could make `ref` abstract and implement stronger-typed overrides
+        # in subclasses.
+        body="""
+            @property
+            def ref(self) -> Any:
+                if self.ref_ is None:
+                    raise ValueError("references not yet resolved on root OME object")
+                return self.ref_()
+        """,
+    ),
+    "BinData": ClassOverride(base_type="object", fields="value: str"),
+    "Map": ClassOverride(fields_suppress={"K"}),
+    "M": ClassOverride(base_type="object", fields="value: str"),
+}
+
+
+def autoflake(text: str, **kwargs: Any) -> str:
+
+    kwargs.setdefault("remove_all_unused_imports", True)
+    kwargs.setdefault("remove_unused_variables", True)
+    return fix_code(text, **kwargs)
+
+
+def black_format(text: str, line_length: int = 88) -> str:
     return black.format_str(text, mode=black.FileMode(line_length=line_length))
 
 
 def sort_imports(text: str) -> str:
-    return isort.SortImports(file_contents=text).output
+    return isort.api.sort_code_string(text, profile="black", float_to_top=True)
 
 
 def sort_types(el: XsdType) -> str:
@@ -251,21 +330,17 @@ def as_identifier(s: str) -> str:
     return _s
 
 
-_CAMEL_SNAKE_OVERRIDES = {"ROIs": "rois"}
+CAMEL_SNAKE_OVERRIDES = {"ROIs": "rois"}
 
 
 def camel_to_snake(name: str) -> str:
-    result = _CAMEL_SNAKE_OVERRIDES.get(name, None)
+    result = CAMEL_SNAKE_OVERRIDES.get(name, None)
     if not result:
-        # FIXME This part must be kept identical to the copy of this function in
-        # the schema module. Ideally we would have one shared implementation but
-        # currently there is a problem importing anything from ome_types if the
-        # model code hasn't been generated yet. It should be fixable with a
-        # little reorganization.
         # https://stackoverflow.com/a/1176023
         result = re.sub("([A-Z]+)([A-Z][a-z]+)", r"\1_\2", name)
         result = re.sub("([a-z0-9])([A-Z])", r"\1_\2", result)
         result = result.lower().replace(" ", "_")
+    camel_snake_registry[name] = result
     return result
 
 
@@ -273,17 +348,38 @@ def local_import(item_type: str) -> str:
     return f"from .{camel_to_snake(item_type)} import {item_type}"
 
 
+def get_docstring(
+    component: Union[XsdComponent, XsdType], summary: bool = False
+) -> str:
+    try:
+        doc = dedent(component.annotation.documentation[0].text).strip()
+        # make sure the first line is followed by a double newline
+        # and preserve paragraphs
+        if summary:
+            doc = re.sub(r"\.\s", ".\n\n", doc, count=1)
+        # textwrap each paragraph seperately
+        paragraphs = ["\n".join(wrap(p.strip(), width=78)) for p in doc.split("\n\n")]
+        # join and return
+        return "\n\n".join(paragraphs)
+    except (AttributeError, IndexError):
+        return ""
+
+
 def make_dataclass(component: Union[XsdComponent, XsdType]) -> List[str]:
+    class_override = CLASS_OVERRIDES.get(component.local_name, None)
     lines = ["from ome_types.dataclasses import ome_dataclass", ""]
-    # FIXME: Refactor to remove BinData special-case.
-    if component.local_name == "BinData":
-        base_type = None
-    elif isinstance(component, XsdType):
+    if isinstance(component, XsdType):
         base_type = component.base_type
     else:
         base_type = component.type.base_type
 
-    if base_type and not hasattr(base_type, "python_type"):
+    if class_override and class_override.base_type:
+        if class_override.base_type == "object":
+            base_name = ""
+        else:
+            base_name = f"({class_override.base_type})"
+        base_type = None
+    elif base_type and not hasattr(base_type, "python_type"):
         base_name = f"({base_type.local_name})"
         if base_type.is_complex():
             lines += [local_import(base_type.local_name)]
@@ -291,25 +387,41 @@ def make_dataclass(component: Union[XsdComponent, XsdType]) -> List[str]:
             lines += [f"from .simple_types import {base_type.local_name}"]
     else:
         base_name = ""
+    if class_override and class_override.imports:
+        lines.append(class_override.imports)
 
     base_members = set()
     _basebase = base_type
     while _basebase:
         base_members.update(set(iter_members(base_type)))
         _basebase = _basebase.base_type
+    skip_names = set()
+    if class_override:
+        skip_names.update(class_override.fields_suppress)
 
-    members = MemberSet(m for m in iter_members(component) if m not in base_members)
+    members = MemberSet(
+        m
+        for m in iter_members(component)
+        if m not in base_members and m.local_name not in skip_names
+    )
     lines += members.imports()
     lines += members.locals()
 
     cannot_have_required_args = base_type and members.has_non_default_args()
     if cannot_have_required_args:
         lines[0] += ", EMPTY"
+    if members.has_nonref_id():
+        lines[0] += ", AUTO_SEQUENCE"
 
     lines += ["@ome_dataclass", f"class {component.local_name}{base_name}:"]
-    # FIXME: Refactor to remove BinData special-case.
-    if component.local_name == "BinData":
-        lines.append("    value: str")
+    doc = get_docstring(component, summary=True)
+    doc = MemberSet(iter_members(component)).docstring(
+        doc or f"{component.local_name}."
+    )
+    doc = f'"""{doc.strip()}\n"""\n'
+    lines += indent(doc, "    ").splitlines()
+    if class_override and class_override.fields:
+        lines.append(class_override.fields)
     lines += members.lines(
         indent=1,
         force_defaults=" = EMPTY  # type: ignore"
@@ -317,22 +429,33 @@ def make_dataclass(component: Union[XsdComponent, XsdType]) -> List[str]:
         else None,
     )
 
+    if class_override and class_override.body:
+        lines.append(class_override.body)
     lines += members.body()
 
     return lines
 
 
-def make_enum(component: XsdComponent, name: str = None) -> List[str]:
-    name = name or component.local_name
+def make_enum(component: XsdComponent) -> List[str]:
+    name = component.local_name
+    _type = component.type if hasattr(component, "type") else component
+    if _type.is_list():
+        _type = _type.item_type
     lines = ["from enum import Enum", ""]
     lines += [f"class {name}(Enum):"]
-    enum_elems = list(component.elem.iter("enum"))
-    facets = component.get_facet(qnames.XSD_ENUMERATION)
+    doc = get_docstring(component, summary=True)
+    if doc:
+        if not doc.endswith("."):
+            doc += "."
+        doc = f'"""{doc}\n"""\n'
+        lines += indent(doc, "    ").splitlines()
+    enum_elems = list(_type.elem.iter("enum"))
+    facets = _type.get_facet(qnames.XSD_ENUMERATION)
     members: List[Tuple[str, str]] = []
     if enum_elems:
         for el, value in zip(enum_elems, facets.enumeration):
             _name = el.attrib["enum"]
-            if component.base_type.python_type.__name__ == "str":
+            if _type.base_type.python_type.__name__ == "str":
                 value = f'"{value}"'
             members.append((_name, value))
     else:
@@ -378,11 +501,12 @@ def iter_members(
         yield from iter_all_members(component)
 
 
+def is_enum_type(obj: XsdType) -> bool:
+    """Return true if XsdType represents an enumeration."""
+    return obj.get_facet(qnames.XSD_ENUMERATION) is not None
+
+
 class Member:
-
-    # Stores plurals from all Members for later access.
-    plurals_registry: Dict[Tuple[str, str], str] = {}
-
     def __init__(self, component: Union[XsdElement, XsdAttribute]):
         self.component = component
         assert not component.is_global()
@@ -394,12 +518,11 @@ class Member:
         name = camel_to_snake(self.component.local_name)
         if self.plural:
             plural = camel_to_snake(self.plural)
-            Member.plurals_registry[(self.parent_name, name)] = plural
+            plural_registry[(self.parent_name, name)] = plural
             name = plural
-        ident = camel_to_snake(name)
-        if not ident.isidentifier():
+        if not name.isidentifier():
             raise ValueError(f"failed to make identifier of {self!r}")
-        return ident
+        return name
 
     @property
     def plural(self) -> Optional[str]:
@@ -421,9 +544,12 @@ class Member:
     def type(self) -> XsdType:
         return self.component.type
 
-    @property
-    def is_enum_type(self) -> bool:
-        return self.type.get_facet(qnames.XSD_ENUMERATION) is not None
+    def to_numpydoc_param(self) -> Parameter:
+        _type = self.type_string
+        _type += ", optional" if self.is_optional else ""
+        desc = get_docstring(self.component)
+        desc = re.sub(r"\s?\[.+\]", "", desc)  # remove bracketed types
+        return Parameter(self.identifier, _type, wrap(desc))
 
     @property
     def is_builtin_type(self) -> bool:
@@ -435,6 +561,31 @@ class Member:
             self.component.schema.builtin_types()["decimal"]
         )
 
+    @property
+    def is_nonref_id(self) -> bool:
+        """Return True for 'id' fields that aren't part of a Reference type."""
+        if self.identifier != "id":
+            return False
+        # Walk up the containment tree until we find something with a base_type.
+        p = self.component.parent
+        while p is not None and not hasattr(p, "base_type"):
+            p = p.parent
+        if p is not None:
+            # Walk the type hierarchy looking for 'Reference'.
+            pt = p.base_type
+            while pt is not None:
+                if pt.local_name == "Reference":
+                    return False
+                pt = pt.base_type
+        # If we get here, we have an 'id' that isn't in a Reference type.
+        return True
+
+    @property
+    def is_ref_id(self) -> bool:
+        if self.identifier == "id":
+            return not self.is_nonref_id
+        return False
+
     @property
     def parent_name(self) -> str:
         """Local name of component's first named ancestor."""
@@ -450,51 +601,51 @@ class Member:
             return self.component.local_name
         return name
 
-    def locals(self) -> Set[str]:
+    def locals(self) -> List[str]:
         if self.key in OVERRIDES:
-            return set()
+            return []
         if isinstance(self.component, (XsdAnyElement, XsdAnyAttribute)):
-            return set()
+            return []
         if not self.type or self.type.is_global():
-            return set()
-        locals_: Set[str] = set()
+            return []
+        locals_: List[str] = []
         # FIXME: this bit is mostly hacks
         if self.type.is_complex() and self.component.ref is None:
-            locals_.add("\n".join(make_dataclass(self.component)) + "\n")
-        if self.type.is_restriction() and self.is_enum_type:
-            locals_.add(
-                "\n".join(make_enum(self.type, name=self.component.local_name)) + "\n"
-            )
+            locals_.append("\n".join(make_dataclass(self.component)) + "\n")
+        if self.type.is_restriction() and is_enum_type(self.type):
+            locals_.append("\n".join(make_enum(self.component)) + "\n")
+        if self.type.is_list() and is_enum_type(self.type.item_type):
+            locals_.append("\n".join(make_enum(self.component)) + "\n")
         return locals_
 
-    def imports(self) -> Set[str]:
+    def imports(self) -> List[str]:
         if self.key in OVERRIDES:
             _imp = OVERRIDES[self.key].imports
-            return set([_imp]) if _imp else set()
+            return [_imp] if _imp else []
         if isinstance(self.component, (XsdAnyElement, XsdAnyAttribute)):
-            return set(["from typing import Any"])
-        imports = set()
+            return ["from typing import Any"]
+        imports = []
         if not self.max_occurs:
-            imports.add("from typing import List")
+            imports.append("from typing import List")
             if self.is_optional:
-                imports.add("from dataclasses import field")
+                imports.append("from dataclasses import field")
         elif self.is_optional:
-            imports.add("from typing import Optional")
+            imports.append("from typing import Optional")
         if self.is_decimal:
-            imports.add("from typing import cast")
+            imports.append("from typing import cast")
         if self.type.is_datetime():
-            imports.add("from datetime import datetime")
+            imports.append("from datetime import datetime")
         if not self.is_builtin_type and self.type.is_global():
             # FIXME: hack
             if not self.type.local_name == "anyType":
                 if self.type.is_complex():
-                    imports.add(local_import(self.type.local_name))
+                    imports.append(local_import(self.type.local_name))
                 else:
-                    imports.add(f"from .simple_types import {self.type.local_name}")
+                    imports.append(f"from .simple_types import {self.type.local_name}")
 
         if self.component.ref is not None:
             if self.component.ref.local_name not in OVERRIDES:
-                imports.add(local_import(self.component.ref.local_name))
+                imports.append(local_import(self.component.ref.local_name))
 
         return imports
 
@@ -521,7 +672,7 @@ class Member:
 
         if self.type.is_global():
             return self.type.local_name
-        elif self.type.is_complex():
+        elif self.type.is_complex() or self.type.is_list():
             return self.component.local_name
 
         if self.type.is_restriction():
@@ -552,7 +703,9 @@ class Member:
         if self.key in OVERRIDES:
             default = OVERRIDES[self.key].default
             return f" = {default}" if default else ""
-        if not self.is_optional:
+        elif self.is_nonref_id:
+            return " = AUTO_SEQUENCE  # type: ignore"
+        elif not self.is_optional:
             return ""
 
         if not self.max_occurs:
@@ -560,7 +713,7 @@ class Member:
         else:
             default_val = self.component.default
             if default_val is not None:
-                if self.is_enum_type:
+                if is_enum_type(self.type):
                     default_val = f"{self.type_string}('{default_val}')"
                 elif hasattr(builtins, self.type_string):
                     default_val = repr(getattr(builtins, self.type_string)(default_val))
@@ -572,13 +725,14 @@ class Member:
 
     @property
     def max_occurs(self) -> bool:
-        return getattr(self.component, "max_occurs", 1)
+        default = None if self.type.is_list() else 1
+        return getattr(self.component, "max_occurs", default)
 
     @property
     def is_optional(self) -> bool:
         # FIXME: hack.  doesn't fully capture the restriction
-        if self.identifier == "id":
-            return True
+        if self.is_ref_id:
+            return False
         if getattr(self.component.parent, "model", "") == "choice":
             return True
         if hasattr(self.component, "min_occurs"):
@@ -598,13 +752,17 @@ class Member:
 
 class MemberSet:
     def __init__(self, initial: Iterable[Member] = ()):
-        self._members: Set[Member] = set()
+        # Use a list to maintain insertion order.
+        self._members: List[Member] = []
         self.update(initial)
 
     def add(self, member: Member) -> None:
         if not isinstance(member, Member):
             member = Member(member)
-        self._members.add(member)
+        # We don't expect very many elements so this O(n) check is fine.
+        if member in self._members:
+            return
+        self._members.append(member)
 
     def update(self, members: Iterable[Member]) -> None:
         for member in members:
@@ -621,23 +779,20 @@ class MemberSet:
         return lines
 
     def imports(self) -> List[str]:
-        if self._members:
-            return list(set.union(*[m.imports() for m in self._members]))
-        return []
+        return list(chain.from_iterable(m.imports() for m in self._members))
 
     def locals(self) -> List[str]:
-        if self._members:
-            return list(set.union(*[m.locals() for m in self._members]))
-        return []
+        return list(chain.from_iterable(m.locals() for m in self._members))
 
     def body(self) -> List[str]:
-        if self._members:
-            return [m.body() for m in self._members]
-        return []
+        return [m.body() for m in self._members]
 
     def has_non_default_args(self) -> bool:
         return any(not m.default_val_str for m in self._members)
 
+    def has_nonref_id(self) -> bool:
+        return any(m.is_nonref_id for m in self._members)
+
     @property
     def non_defaults(self) -> "MemberSet":
         return MemberSet(m for m in self._members if not m.default_val_str)
@@ -645,6 +800,13 @@ class MemberSet:
     def __iter__(self) -> Iterator[Member]:
         return iter(self._members)
 
+    def docstring(self, summary: str = "") -> str:
+        ds = NumpyDocString(summary)
+        ds["Parameters"] = [
+            m.to_numpydoc_param() for m in sorted(self._members, key=sort_prop)
+        ]
+        return str(ds)
+
 
 class GlobalElem:
     def __init__(self, elem: Union[XsdElement, XsdType]):
@@ -735,8 +897,7 @@ class GlobalElem:
         return lines
 
     def lines(self) -> str:
-        # FIXME: Refactor to remove BinData special-case.
-        if not self.is_complex and self.elem.local_name != "BinData":
+        if not self.is_complex:
             lines = self._simple_class()
         elif self.elem.abstract:
             lines = self._abstract_class()
@@ -745,7 +906,7 @@ class GlobalElem:
         return "\n".join(lines)
 
     def format(self) -> str:
-        return black_format(sort_imports(self.lines() + "\n"))
+        return black_format(sort_imports(autoflake(self.lines() + "\n")))
 
     def write(self, filename: str) -> None:
         os.makedirs(os.path.dirname(filename), exist_ok=True)
@@ -758,8 +919,7 @@ class GlobalElem:
 
 
 _this_dir = os.path.dirname(__file__)
-# _url = os.path.join(_this_dir, "ome_types", "ome-2016-06.xsd")
-_url = "https://www.openmicroscopy.org/Schemas/OME/2016-06/ome.xsd"
+_url = os.path.join(_this_dir, "ome_types", "ome-2016-06.xsd")
 _target = os.path.join(_this_dir, "ome_types", "model")
 
 
@@ -801,10 +961,27 @@ def convert_schema(url: str = _url, target_dir: str = _target) -> None:
         text += local_import(classname) + "\n"
     text = sort_imports(text)
     text += f"\n\n__all__ = [{', '.join(sorted(repr(i[1]) for i in init_imports))}]"
-    # FIXME This could probably live somewhere else less visible to end-users.
-    text += "\n\n_field_plurals = " + repr(Member.plurals_registry)
+    # FIXME These could probably live somewhere else less visible to end-users.
+    if len(plural_registry) != len(set(plural_registry.values())):
+        raise Exception(
+            "singular-to-plural mapping is not invertible (duplicate plurals)"
+        )
+    text += "\n\n_singular_to_plural = " + repr(
+        {k: plural_registry[k] for k in sorted(plural_registry)}
+    )
+    text += "\n\n_plural_to_singular = " + repr(
+        {plural_registry[k]: k[1] for k in sorted(plural_registry)}
+    )
+    if len(camel_snake_registry) != len(set(camel_snake_registry.values())):
+        raise Exception("camel-to-snake mapping is not invertible (duplicate snakes)")
+    text += "\n\n_camel_to_snake = " + repr(
+        {k: camel_snake_registry[k] for k in sorted(camel_snake_registry)}
+    )
+    text += "\n\n_snake_to_camel = " + repr(
+        {camel_snake_registry[k]: k for k in sorted(camel_snake_registry)}
+    )
     text = black_format(text)
-    with open(os.path.join(target_dir, f"__init__.py"), "w") as f:
+    with open(os.path.join(target_dir, "__init__.py"), "w") as f:
         f.write(text)
 
 
